#include "DataStructures.h"

// temp code for Add Point redo
//#include "RootflyDlg.h"

/*the mother class of all command classes*/

class Command 
{ 
public: 
  virtual bool Execute() = 0; 
  virtual bool Unexecute() = 0;  
};

// The following derived classes correspond to the possible
// "undoable" actions.

// -------------------------------------------------

class UC_ResizeDiameter : public Command
{
public:
	UC_ResizeDiameter(RootflyData *data, int index, int root_id, double old_radius, double new_radius) {
		m_index = index;
		m_root_id = root_id;
		m_data = data;
		m_new_radius = new_radius;
		m_old_radius = old_radius;
	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData *m_data;
	int m_index;
	int m_root_id;
	double m_new_radius;
	double m_old_radius;
};


// -------------------------------------------------

class UC_SetRootAlive : public Command
{
public:
	UC_SetRootAlive(RootflyData *data, int rootid, bool alive) 
	{
		m_data=data; 
		m_rootid=rootid;
		m_alive=alive;

		// need to grab the session that we were on when this was
		// originally called
		ApplicationData& app = m_data->GetApplicationData();
		m_session = app.GetCurrentSession();
	}

	bool Execute();
	bool Unexecute();
	
private:
	RootflyData *m_data;
	int m_rootid;
	bool m_alive;
	int m_session;
};


// -------------------------------------------------
class UC_MoveRootSess: public Command
{
public:
  UC_MoveRootSess(RootflyData *data, int root_id, int session, int offset_x, int offset_y) {
	  m_data = data;
	  m_root_id = root_id;
	  m_session = session;
    m_offset_x = offset_x;
	  m_offset_y = offset_y;
	  m_executed_once = 0;
	}
  bool Execute();
	bool Unexecute();

private:
	bool m_executed_once;  // Needed because of offset storage method

	RootflyData *m_data;
	int m_root_id;
  int m_session;
	
  double m_offset_x;
	double m_offset_y;
  
  RootInstance *m_root_inst;

  // Internal functions to get root instance, move diameter and
  // move point. GetRootInstance has been created to optimize the code
  // in the Execute() and Unexecute() functions. 
  
  // Move Diameter and Move Point have been defined as internal functions. 
  // Their versions from DataStructures may be directly used without losing
  // any functionality.

  RootInstance *RootSessGetRootInstance( int m_root_id )
  {
    Experiment& exp = m_data->GetExperiment();
    Tube *tube = exp.GetTubeByName(m_data->GetApplicationData().GetCurrentTubeName());

    /// Failure if tube is null
    if( tube == NULL ) return false;
  
    Root *r_root = tube->GetRoot(m_root_id);
    /// Failure if root is null
    if( r_root == NULL ) return false;
  
    return r_root->GetRootInstance(m_session);
  } 

  bool RootSessMoveDiameter(RootInstance *root_inst, int index, int root_id, int x, int y)
  {  
    if(root_inst->MoveDiameter(index, x, y)) return true;
    else return false;
  }

  bool RootSessMovePoint(RootInstance *root_inst, int index, int root_id, int x, int y)
  { 
    /// Move the point
    if(root_inst->MovePoint(index, x, y)) return true;
    else return false;  
  }

};


// -------------------------------------------------
class UC_MoveDiameter: public Command
{

public:
  UC_MoveDiameter(RootflyData *data, int index, int root_id, const RPoint& old_r, const RPoint& new_r){
    m_data = data;
    m_index = index;
    m_root_id = root_id;
    m_new_x = new_r.x;
    m_new_y = new_r.y;
    m_new_window = new_r.window;
	m_old_x = old_r.x;
	m_old_y = old_r.y;
	m_old_window = old_r.window;
  }
  bool Execute();
	bool Unexecute();

private:
  RootflyData *m_data;
  int m_index;
  int m_root_id;
  int m_new_x;
  int m_new_y;
  int m_new_window;
  int m_old_x;
  int m_old_y;
  int m_old_window;
  RootInstance *m_root_inst;

  // GetRootInstance has been created to optimize the code
  // in the Execute() and Unexecute() functions. 
  
  RootInstance *MoveDiameterGetRootInstance (int m_root_id )
  {

    /// Get the current tube and session number from the ApplicationData
    ApplicationData& app_data = m_data->GetApplicationData();
    CString tube_tag = app_data.GetCurrentTubeName();
    int session = app_data.GetCurrentSession();
  
    /// Get the tube from the experiment using the tag from the ApplicationData
    Experiment& exp = m_data->GetExperiment();
    Tube* tube = exp.GetTubeByName(tube_tag);
    
     /// Failure if tube is null
    if(tube == NULL)  return NULL;
      
    /// Get the root from the tube by the Root ID
    Root* temp_root = tube->GetRoot(m_root_id);
    
    /// Failure if root is null
    if(temp_root == NULL) return NULL;
    
    /// Get the root instance using the session number from ApplicationData
    return temp_root->GetRootInstance(session);
  }
};


// -------------------------------------------------
class UC_MoveRootsInWindow: public Command
{

public:
  UC_MoveRootsInWindow(RootflyData *data,int window,int offset_x, int offset_y)
  {
	executed_once = 0;  // Needed because of offset storage method
    m_data = data;
    m_window = window;
    m_offset_x = offset_x;
    m_offset_y = offset_y;
  }	

  bool Execute();
	bool Unexecute();


private:
  bool executed_once;
  RootflyData *m_data;
  int m_window;
  int m_offset_x;
  int m_offset_y;
  int m_session;
  int m_root_id;
  RootInstance *m_root_inst;

  // Internal functions to get root instance, move root session, move diameter and
  // move point. GetRootInstance has been created to optimize the code
  // in the Execute() and Unexecute() functions. 
  
  // Move Diameter and Move Point have been defined as internal functions. 
  // Their versions from DataStructures may be directly used without losing
  // any functionality.

  bool RootWindowMoveRootSess(int m_root_id, int session, int m_offset_x, int m_offset_y )
  {
    m_session = session;
    m_root_inst = RootWindowGetRootInstance( m_root_id );
    if( m_root_inst == NULL ) return false;
  
    /// For all points
    for(int i = 0;i<m_root_inst->GetAllPoints()->size();i++)
     RootWindowMovePoint(m_root_inst, i, m_root_id, m_root_inst->GetPoint(i)->x+m_offset_x,m_root_inst->GetPoint(i)->y+m_offset_y);//may need to update this to be a bit different with x and y
  
    /// Move Diameter (assumes only one diameter)
    RootWindowMoveDiameter(m_root_inst, 0,m_root_id, m_root_inst->GetDiameter(0)->r_point.x + m_offset_x, m_root_inst->GetDiameter(0)->r_point.y + m_offset_y);
  
    return true;
  }

  RootInstance *RootWindowGetRootInstance( int m_root_id )
  {
    Experiment& exp = m_data->GetExperiment();
    Tube *tube = exp.GetTubeByName(m_data->GetApplicationData().GetCurrentTubeName());

    /// Failure if tube is null
    if( tube == NULL ) return false;
  
    Root *r_root = tube->GetRoot(m_root_id);
    /// Failure if root is null
    if( r_root == NULL ) return false;
  
    return r_root->GetRootInstance(m_session);
  } 


  bool RootWindowMoveDiameter(RootInstance *root_inst, int index, int root_id, int x, int y)
  {  
    if(root_inst->MoveDiameter(index, x, y)) return true;
    else return false;
  }

  bool RootWindowMovePoint(RootInstance *root_inst, int index, int root_id, int x, int y)
  { 
    /// Move the point
    if(root_inst->MovePoint(index, x, y)) return true;
    else return false;  
  }
  
};



 
// -------------------------------------------------

class UC_SetRootColor : public Command
{
public:
	UC_SetRootColor(RootflyData *data, int root_id, int color_choice) 
	{
		m_data=data; 
		m_root_id=root_id;
		m_color_choice=color_choice;
	}

	bool Execute();
	bool Unexecute();
	
private:
	RootflyData *m_data;
	int m_root_id;
	int m_color_choice;
	int m_old_color;
};


// -------------------------------------------------

class UC_MovePoint : public Command
{
public:
	UC_MovePoint(RootflyData *data, int index, int root_id, int new_x, int new_y, int old_x, int old_y, int new_diam_x, int new_diam_y, int old_diam_x, int old_diam_y, int new_window, int old_window) 
	{
		m_data=data; 
		m_root_id=root_id;
		m_new_x = new_x;
		m_new_y = new_y;
		m_old_x = old_x;
		m_old_y = old_y;
		m_index = index;
		m_new_diam_x = new_diam_x;
		m_new_diam_y = new_diam_y;
		m_old_diam_x = old_diam_x;
		m_old_diam_y = old_diam_y;
		m_new_window = new_window;
		m_old_window = old_window;

		executed_once = false;
	}

	bool Execute();
	bool Unexecute();
	
private:
	RootflyData *m_data;

	bool executed_once;

	int m_root_id;
	int m_index;

	int m_new_x;
	int m_new_y;
	int m_old_x;
	int m_old_y;
	int m_new_window;
	int m_old_window;

  int m_old_diam_x;
  int m_old_diam_y;
  int m_new_diam_x;
  int m_new_diam_y;

  RootInstance *m_root_inst;


	RootInstance *MovePointGetRootInstance(int m_root_id)
	{
	/// Get the current tube and session number from the ApplicationData
    ApplicationData& app_data = m_data->GetApplicationData();
    CString tube_tag = app_data.GetCurrentTubeName();
    int session = app_data.GetCurrentSession();
  
    /// Get the tube from the experiment using the tag from the ApplicationData
    Experiment& exp = m_data->GetExperiment();
    Tube* tube = exp.GetTubeByName(tube_tag);
    
     /// Failure if tube is null
    if(tube == NULL)  return NULL;
      
    /// Get the root from the tube by the Root ID
    Root* temp_root = tube->GetRoot(m_root_id);
    
    /// Failure if root is null
    if(temp_root == NULL) return NULL;
    
    /// Get the root instance using the session number from ApplicationData
    return temp_root->GetRootInstance(session);
	}

};


// -------------------------------------------------
class UC_AddPoint : public Command
{

public:
  UC_AddPoint(RootflyData *data,int root_id,int index)
  {
    m_data = data;
    m_root_id = root_id;
    m_index = index; 
   
    // initialize temp pt such that one can infer if
    // this is the first execution or a redo
    m_temp_point.x = -1;
    m_temp_point.y = -1;
    m_temp_point.window = -1;
 }

  bool Execute();
	bool Unexecute();

 
private:
  RootflyData *m_data;
  int m_root_id;
  int m_index;
  RootInstance *m_root_inst;

  RDiameter *m_diameter;
  int m_diam_x;
  int m_diam_y;

  RPoint *m_point;
  RPoint m_temp_point;
  

  RootInstance *AddPointGetRootInstance( int m_root_id )
  {
    /// Get the current tube and session number from the ApplicationData
    ApplicationData& app_data = m_data->GetApplicationData();
    CString tube_tag = app_data.GetCurrentTubeName();
    int session = app_data.GetCurrentSession();
  
    /// Get the tube from the experiment using the tag from the ApplicationData
    Experiment& exp = m_data->GetExperiment();
    Tube* tube = exp.GetTubeByName(tube_tag);
    
     /// Failure if tube is null
    if(tube == NULL)  return NULL;
      
    /// Get the root from the tube by the Root ID
    Root* temp_root = tube->GetRoot(m_root_id);
    
    /// Failure if root is null
    if(temp_root == NULL) return NULL;
    
    /// Get the root instance using the session number from ApplicationData
    return temp_root->GetRootInstance(session);
  } 

};


// -------------------------------------------------
class UC_DeletePoint : public Command
{

public:
  UC_DeletePoint(RootflyData *data,int root_id,int index)
  {
    m_data = data;
    m_root_id = root_id;
    m_index = index; 

    // initialize temp pt such that one can infer if
    // this is the first execution or a redo
    m_temp_point.x = -1;
    m_temp_point.y = -1;
    m_temp_point.window = -1;

  }

  bool Execute();
	bool Unexecute();

  
private:
  RPoint *m_point;
  RPoint m_temp_point;

  RDiameter *m_diameter;
  int m_diam_x;
  int m_diam_y;

  RootflyData *m_data;
  int m_root_id;
  int m_index;
  RootInstance *m_root_inst;


  RootInstance *DeletePointGetRootInstance( int m_root_id )
  {
    /// Get the current tube and session number from the ApplicationData
    ApplicationData& app_data = m_data->GetApplicationData();
    CString tube_tag = app_data.GetCurrentTubeName();
    int session = app_data.GetCurrentSession();
  
    /// Get the tube from the experiment using the tag from the ApplicationData
    Experiment& exp = m_data->GetExperiment();
    Tube* tube = exp.GetTubeByName(tube_tag);
    
     /// Failure if tube is null
    if(tube == NULL)  return NULL;
      
    /// Get the root from the tube by the Root ID
    Root* temp_root = tube->GetRoot(m_root_id);
    
    /// Failure if root is null
    if(temp_root == NULL) return NULL;
    
    /// Get the root instance using the session number from ApplicationData
    return temp_root->GetRootInstance(session);
  } 

};

// -------------------------------------------------

class UC_CreateRoot: public Command
{
public:
	UC_CreateRoot(RootflyData* data, int primary_window, const RootInstance& ri_orig, CString default_cond) {

		ApplicationData& app_data = data->GetApplicationData();
		m_tube_tag = app_data.GetCurrentTubeName();  
		m_session = app_data.GetCurrentSession();
  
		m_data = data;
  		m_ri_orig = ri_orig;
		m_primary_window = primary_window;
		m_default_cond = default_cond;
		m_executed = false;
		m_root_id = 0;

	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData *m_data;
	CString m_tube_tag;
	RootInstance m_ri_orig;
	int m_root_id;
	int m_session;
	int m_primary_window;
	CString m_default_cond;
	bool m_executed;
};

// -------------------------------------------------

class UC_DeleteRoot: public Command
{
public:
	UC_DeleteRoot(RootflyData* data, CString tube_tag, int root_id, Root* root_orig) {

		m_data = data;
		m_tube_tag = tube_tag;
		m_root_id = root_id;
		m_root_orig = *(root_orig);
	    m_primary_window = root_orig->GetPrimaryWindow();

	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData* m_data;
	CString m_tube_tag;
	int m_root_id;
	Root m_root_orig;
	int m_primary_window;
};

// -------------------------------------------------

class UC_CreatePoint : public Command
{
public:
	UC_CreatePoint(RootInstance *root_inst, int root_id){
		m_root_inst = root_inst;
		m_root_id = root_id;
	}
	bool Execute();
	bool Unexecute();

	RPoint *m_temp_point;

private:
	RootInstance *m_root_inst;
	int m_root_id;
};

// ---------------------------------------------------

class UC_CopySession : public Command
{
public:
	UC_CopySession(RootflyData *data, int from, int to, const CString &tubename, int curr_window){
		m_data = data;
		m_from = from;
		m_to = to;
		m_tubename = tubename;
		m_curr_window = curr_window;
	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData *m_data;
	int m_from;
	int m_to;
	vector<Root> *m_roots;
	CString m_tubename; 
	int m_curr_window;
};

// ---------------------------------------------------

class UC_ExtendEndpoint : public Command
{
public:
	UC_ExtendEndpoint(RootflyData* data, RootInstance &old_ri, int root_id, int session)
	{
		m_data = data;
		m_old_ri = old_ri;
		m_root_id = root_id;
		m_session = session;
		executed = false;
	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData *m_data;
	RootInstance m_old_ri;
	RootInstance m_new_ri;
	int m_root_id;
	int m_session;
	bool executed;
};

// ---------------------------------------------------

class UC_SetConditionCode : public Command
{
public:
	UC_SetConditionCode(RootflyData* data, int root_id, CString condition)
	{
		m_data = data;
		m_root_id = root_id;
		m_condition = condition;
	}
	bool Execute();
	bool Unexecute();

private:
	RootflyData *m_data;
	int m_root_id;
	CString m_condition;
	CString m_prev_condition;
};