// TubeControl.cpp : implementation file
//

#include "stdafx.h"
#include "rootfly.h"
#include "TubeControl.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// TubeControl

TubeControl::TubeControl(ProgressData *data) : m_control_size(0, 0)
{
  m_data = data;
  m_session = 1;
  
  COLOR_NOT_EXIST = RGB(200,200,200);
  COLOR_NOT_ANNOTATED = RGB(255,0,0);
  COLOR_NOT_DONE = RGB(255,255,0);
  COLOR_DONE = RGB(0,255,0);

  m_control_margin = m_tube_spacing = 15;
  m_tube_margin = m_window_spacing = 3;
  m_tube_top = 10;
  m_window_size = CSize(20, 15);
  m_row_length = 1;
  m_column_length = 2;
  m_label_height = 12;
  
  while (m_row_length * m_column_length < data->GetWindowCount())
  {
    if ((float) m_column_length < (float) m_row_length * 1.5)
    {
      m_column_length++;
    }
    else
    {
      m_row_length++;
    }
  }

  m_tube_size = CSize(m_window_size.cx * m_row_length + (m_row_length - 1) * m_window_spacing + 2 * m_tube_margin,
    m_tube_top * 1.5 + m_window_size.cy * m_column_length + (m_column_length - 1) * m_window_spacing + 2 * m_tube_margin);

  m_tool_tip.Create(this);
}

TubeControl::~TubeControl()
{
}

void TubeControl::OnFinalRelease()
{
	// When the last reference for an automation object is released
	// OnFinalRelease is called.  The base class will automatically
	// deletes the object.  Add additional cleanup required for your
	// object before calling the base class.

	CStatic::OnFinalRelease();
}


BEGIN_MESSAGE_MAP(TubeControl, CStatic)
	//{{AFX_MSG_MAP(TubeControl)
	ON_WM_PAINT()
	ON_WM_LBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BEGIN_DISPATCH_MAP(TubeControl, CStatic)
	//{{AFX_DISPATCH_MAP(TubeControl)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_DISPATCH_MAP
END_DISPATCH_MAP()

// Note: we add support for IID_ITubeControl to support typesafe binding
//  from VBA.  This IID must match the GUID that is attached to the 
//  dispinterface in the .ODL file.

// {D9B9BB98-A043-427E-A6F1-F755D2B1211B}
static const IID IID_ITubeControl =
{ 0xd9b9bb98, 0xa043, 0x427e, { 0xa6, 0xf1, 0xf7, 0x55, 0xd2, 0xb1, 0x21, 0x1b } };

BEGIN_INTERFACE_MAP(TubeControl, CStatic)
	INTERFACE_PART(TubeControl, IID_ITubeControl, Dispatch)
END_INTERFACE_MAP()

/////////////////////////////////////////////////////////////////////////////
// TubeControl message handlers

void TubeControl::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
  
  CRect rect;
  GetClientRect(rect);
  if (m_control_size != rect.Size())
  {
    m_control_size = rect.Size();
    PlaceWindows();
  }

  DrawWindows(&dc);
  DrawLegend(&dc);

	
	// TODO: Add your message handler code here
	
	// Do not call CStatic::OnPaint() for painting messages
}

void TubeControl::SetSession(const int &session)
{
  m_session = session;
}

void TubeControl::SetTube(const CString &tube)
{
  m_tube = tube;
}
TubeControl::WindowRect::WindowRect() : CRect()
{
  window = 0;
}

TubeControl::WindowRect::WindowRect(const CPoint &point, const CSize &size, const int &window) : CRect(point, size)
{
  this->window = window;
}

TubeControl::TubeRect::TubeRect() : CRect()
{
  tube = 0;
  windows = std::vector<WindowRect>();
}

TubeControl::TubeRect::TubeRect(const CPoint &point, const CSize &size, const int &tube) : CRect(point, size)
{
  this->tube = tube;
  windows = std::vector<WindowRect>();
}

void TubeControl::PlaceWindows()
{
  int tube_x, tube_y, win_x, win_y, tube, win, t;

  const std::vector<bool> *tubes = m_data->GetTubeList();
  const std::vector<bool> *wins = m_data->GetWindowList();

  tube = t = -1;

  for (tube_y = m_control_margin; (tube + 1 < tubes->size()) && (tube_y + m_tube_size.cy + m_control_margin + m_label_height <= m_control_size.cy); tube_y += m_tube_size.cy + m_tube_spacing + m_label_height)
  {
    for (tube_x = m_control_margin; (tube + 1 < tubes->size()) && (tube_x + m_tube_size.cx + m_control_margin <= m_control_size.cx); tube_x += m_tube_size.cx + m_tube_spacing)
    {
      do 
      {
        tube++;
      } while ((tube < tubes->size()) && !(tubes->operator[](tube)));
      t++;
      
      if (tube >= tubes->size())
      {
        break;
      }
      
      if (m_tubes.size() <= t)
      {
        m_tubes.resize(t + 1, TubeRect());
      }
      m_tubes[t] = TubeRect(CPoint(tube_x, tube_y), m_tube_size, tube);
      if (m_tubes[t].windows.size() < wins->size())
      {
        m_tubes[t].windows.resize(wins->size(), WindowRect());
      }
      
      win = 0;

      for (win_y = m_tube_margin + m_tube_top; win_y + m_window_size.cy + m_tube_margin <= m_tube_size.cy + m_tube_top; win_y += m_window_size.cy + m_window_spacing)
      {
        for (win_x = m_tube_margin; win_x + m_window_size.cx + m_tube_margin <= m_tube_size.cx; win_x += m_window_size.cx + m_window_spacing)
        {
          do
          {
            win++;
          } while ((win < wins->size()) && !(wins->operator[](win)));

          if (win < wins->size())
          {
            m_tubes[t].windows[win] = WindowRect(CPoint(tube_x + win_x, tube_y + win_y), m_window_size, win);
          }
        }
      }
    }
  }
}

void TubeControl::DrawWindows(CPaintDC *dc)
{
  int t, w;

  for (t = 0; t < m_tubes.size(); t++)
  {
    if (m_tubes[t].Width() == 0)
    {
      continue;
    }

    DrawTube(m_tubes[t], dc);

    for (w = 1; w < m_tubes[t].windows.size(); w++)
    {
      if (m_tubes[t].windows[w].Width() == 0)
      {
        continue;
      }

      DrawWindow(m_tubes[t].windows[w], ChooseColor(m_tubes[t].tube, m_tubes[t].windows[w].window), dc);
    }
  }
}

void TubeControl::DrawTube(const TubeRect &tube, CPaintDC *dc)
{
  dc->SetBkMode(TRANSPARENT);

	//draw the sides of the tube
	dc->MoveTo(tube.left, tube.bottom - m_tube_top / 2);
	dc->LineTo(tube.left, tube.top + m_tube_top / 2);
	dc->MoveTo(tube.right, tube.bottom - m_tube_top / 2);
	dc->LineTo(tube.right, tube.top + m_tube_top / 2);

	//draw the top and bottom of the tube
	CRect top(tube.left, tube.top, tube.right, tube.top + m_tube_top);
	dc->Ellipse(top);

  CRect bottom(tube.left, tube.bottom - m_tube_top, tube.right, tube.bottom);
	dc->Arc(bottom, CPoint(bottom.left, (bottom.top + bottom.bottom) / 2), CPoint(bottom.right, (bottom.top + bottom.bottom) / 2));

  // Describe a 24-point truetype font of normal weight
  LOGFONT lf;
  memset(&lf, 0, sizeof(lf));
  lf.lfHeight = -MulDiv(8, dc->GetDeviceCaps(LOGPIXELSY), 72);
  lf.lfWeight = FW_NORMAL;
  lf.lfOutPrecision = OUT_TT_ONLY_PRECIS;

  // create and select it
  CFont newFont;
  newFont.CreateFontIndirect(&lf);
  CFont* pOldFont = dc->SelectObject(&newFont);

  // create the label text
  CString text;
  text.Format("Tube %d", tube.tube);

  // use a path to record how the text was drawn
  dc->SetTextAlign(TA_CENTER | TA_BASELINE);
  dc->TextOut((tube.left + tube.right) / 2, tube.bottom + m_label_height, _T(text));
}

void TubeControl::DrawWindow(const WindowRect &window, const COLORREF &color, CPaintDC *dc)
{
	//draw a rectangle
	dc->MoveTo(window.left, window.top);
	dc->LineTo(window.left, window.bottom);
	dc->LineTo(window.right, window.bottom);
	dc->LineTo(window.right, window.top);
	dc->LineTo(window.left, window.top);

  //color the rectangle
  dc->FillSolidRect(CRect(window.TopLeft() + CSize(1,1), window.Size() - CSize(1,1)), color);
}

COLORREF TubeControl::ChooseColor(const int &tube, const int &window)
{
  if (!m_data->GetExists(tube, window, m_session))
  {
    return COLOR_NOT_EXIST; //COLOR_NOT_EXIST;
  }
  else if (m_data->GetDone(tube, window, m_session))
  {
    return COLOR_DONE; //COLOR_DONE;
  }
  else if (m_data->GetAnnotated(tube, window, m_session))
  {
    return COLOR_DONE; //COLOR_NOT_DONE;
  }
  else
  {
    return COLOR_NOT_ANNOTATED; //COLOR_NOT_ANNOTATED;
  }
}


void TubeControl::OnLButtonDown(UINT nFlags, CPoint point) 
{
  int tube, window;
  CRect rect;

  GetAt(point, &tube, &window, &rect);

  if ((tube != -1) && (window != -1) && (m_data->GetExists(tube, window, m_session)))
  {
    SelectImage(tube, m_session, window);
  }
	
	CStatic::OnLButtonDown(nFlags, point);
}

void TubeControl::GetAt(CPoint point, int *tube, int *window, CRect *rect)
{
  int t, w;

  for (t = 0; t < m_tubes.size(); t++)
  {
    if (!m_tubes[t].PtInRect(point))
    {
      continue;
    }

    for (w = 0; w < m_tubes[t].windows.size(); w++)
    {
      if (m_tubes[t].windows[w].PtInRect(point))
      {
        *tube = m_tubes[t].tube;
        *window = m_tubes[t].windows[w].window;
        *rect = m_tubes[t].windows[w];
        return;
      }
    }
  }

  *tube = *window = -1;
}

void TubeControl::SetImageSelectCallback(TubeControl::ImageSelectCallback *callback)
{
  m_select_image = callback;
}

// this set up may seem overly complicated, but it does avoid required TubeControl to have knowledge of ProgressDlg
void TubeControl::SelectImage(const int &tube, const int &session, const int &window)
{
  m_select_image->SelectImage(tube, session, window);
}

void TubeControl::DrawLegend(CPaintDC *dc)
{
  dc->MoveTo(500,370);
  dc->LineTo(520,370);
  dc->LineTo(520, 385);
  dc->LineTo(500, 385);
  dc->LineTo(500,370);
  CRect done;
  done.SetRect(500,370,520,385);
  dc->FillSolidRect(done, COLOR_DONE);

  dc->MoveTo(500,390);
  dc->LineTo(520,390);
  dc->LineTo(520, 405);
  dc->LineTo(500, 405);
  dc->LineTo(500,390);
  CRect empty;
  empty.SetRect(500,390,520,405);
  dc->FillSolidRect(empty, COLOR_NOT_ANNOTATED);
  
  dc->MoveTo(500,410);
  dc->LineTo(520,410);
  dc->LineTo(520, 425);
  dc->LineTo(500, 425);
  dc->LineTo(500,410);
  CRect noImage;
  noImage.SetRect(500,410,520,425);
  dc->FillSolidRect(noImage, COLOR_NOT_EXIST);
}

void TubeControl::OnMouseMove(UINT nFlags, CPoint point) 
{
  CRect rect;
  int tube, window;

  GetAt(point, &tube, &window, &rect);
  if ((tube > 0) && (window > 0) && m_data->GetExists(tube, window, m_session))
  {
    CString text;
    text.Format("Window %d", window);
    m_tool_tip.AddTool(this, text);
    m_tool_tip.Activate(true);
  }
  else
  {
    m_tool_tip.Activate(false);
  }

	CStatic::OnMouseMove(nFlags, point);
}

BOOL TubeControl::PreTranslateMessage(MSG* pMsg) 
{
  m_tool_tip.RelayEvent(pMsg);
	
	return CStatic::PreTranslateMessage(pMsg);
}
